#include "Match.h"

using namespace std;

Match* Match::matchInstance = 0;

Match::Match() {
  setupDictionary(dictionary);
}

Match* Match::instance() {
  if (!matchInstance)
    matchInstance = new Match;
  return matchInstance;
}

// If word_list.txt is missing the dictionary will be empty
void Match::setupDictionary(boost::unordered_set<string>& dictionary) {
  ifstream inf("word_list.txt");
  if (!inf) {
    cerr << "word_list.txt could not be opened for reading!" << endl;
    return;
  }
  dictionary.insert(std::istream_iterator<std::string>(inf), std::istream_iterator<std::string>());
}

const boost::unordered_set<string>& Match::getDictionary() {
  return dictionary;
}

string Match::getRandomWordFromDictionary() {
  int numWords = dictionary.size();
  static boost::mt19937 rng(time(0));
  boost::uniform_int<> zero_to_numwords(0, numWords);
  boost::variate_generator<boost::mt19937&, boost::uniform_int<> > randomWordNumber(rng, zero_to_numwords);
  boost::unordered_set<string>::iterator it = dictionary.begin();
  advance(it, randomWordNumber());
  return *it;
}

vector<string> Match::keyVectorFromMap(const boost::unordered_map<string, vector<string> >& m) {
  vector<string> v;
  pair<string, vector<string> > p;
  BOOST_FOREACH (p, m) {
    v.push_back(p.first);
  }
  return v;
}

vector<string> Match::match(const boost::unordered_map<string, vector<string> >& m1, const boost::unordered_map<string, vector<string> >& m2) {
  // see if there are any words in common in the key values of the two maps
  // return zero sized vector if nothing is in common; otherwise, return the path from the start word to the finish word
  vector<string> keyVector1 = keyVectorFromMap(m1);
  vector<string> keyVector2 = keyVectorFromMap(m2);
  vector<string>::iterator it = find_first_of(keyVector1.begin(), keyVector1.end(), keyVector2.begin(), keyVector2.end());
  vector<string> v;
  if (it == keyVector1.end()) {
    return v;
  } else {
    // m1's value contains a list of words from the starting word to the word in common
    // m2's value contains a list of words from the ending word to the word in common
    // from these we make a list from the starting word to the ending word
    vector<string> valueVector1 = m1.find(*it)->second;
    vector<string> valueVector2 = m2.find(*it)->second;

    BOOST_FOREACH (string s, valueVector1) {
      v.push_back(s);
    }
    vector<string>::reverse_iterator rit = ++valueVector2.rbegin();
    for ( ; rit != valueVector2.rend(); ++rit) {
      v.push_back(*rit);
    }
    return v;
  }
}

vector<string> Match::shortestPathFromStartWordToEndWord(string startWord, string endWord) {
  startWord = trimWord(startWord);
  endWord = trimWord(endWord);
  convertToLowerCase(startWord);
  convertToLowerCase(endWord);
  WordSearch startWordSearch(startWord, dictionary);
  WordSearch endWordSearch(endWord, dictionary);
  bool isStartTurn = true;
  vector<string> matchPath;
  int numberWordsEverUsed;
  while ((matchPath = match(startWordSearch.getWordsCurrent(), endWordSearch.getWordsCurrent())).size() == 0) {
    // alternate from which end to find the next level of words in the dictionary
    WordSearch* pWordSearch;
    if (isStartTurn)
      pWordSearch = &startWordSearch;
    else
      pWordSearch = &endWordSearch;

    // check to see whether a dead end has been reached
    numberWordsEverUsed = pWordSearch->getNumberWordsEverUsed();
    pWordSearch->allPermutations();
    if (numberWordsEverUsed == pWordSearch->getNumberWordsEverUsed()) {

      vector<string> v;
      return v;
    }
    isStartTurn = !isStartTurn;
  }

  return matchPath;
}

string Match::trimWord(string word) {
  while (word[0] == ' ' || word[0] == '\t')
    word.erase(0, 1);
  while (word[word.length() - 1] == ' ' || word[word.length() - 1] == '\t')
    word.erase(word.length() - 1, 1);
  return word;
}

void Match::convertToLowerCase(string& word) {
  std::transform(word.begin(), word.end(), word.begin(), ::tolower);
}

string Match::validateInput(string startWord, string endWord) {
  startWord = trimWord(startWord);
  endWord = trimWord(endWord);
  convertToLowerCase(startWord);
  convertToLowerCase(endWord);
  string result;
  // make sure the string contains only one word
  if (startWord.length() == 0) {
    result = "Please enter starting word.";
    return result;
  }
  if (endWord.length() == 0) {
    result = "Please enter ending word.";
    return result;
  }
  if (startWord.find_first_of(" \t") != string::npos) {
    result = "More than one word entered for starting word.";
    return result;
  }
  if (endWord.find_first_of(" \t") != string::npos) {
    result = "More than one word entered for ending word.";
    return result;
  }
  static const boost::regex e("[a-z]*");
  if (!boost::regex_match(startWord, e)) {
    result = "Your starting word contains non-letter characters.";
    return result;
  }
  if (!boost::regex_match(endWord, e)) {
    result = "Your ending word contains non-letter characters.";
    return result;
  }
  //make sure the starting and ending words are in the dictionary
  if (dictionary.find(startWord) == dictionary.end()) {
    result = "Your starting word, \"" + startWord + "\", is not in the dictionary.";
    return result;
  }
  if (dictionary.find(endWord) == dictionary.end()) {
    result = "Your ending word, \"" + endWord + "\", is not in the dictionary.";
    return result;
  }
  return result;
}
